from .wrapper import Wrapper
from .inputs import get_inputs
from .embedding import Embeddings
from .attention import Attention
from .multi_head import MultiHeadAttention
from .feed_forward import FeedForward
from .layer_norm import LayerNormalization
from .transformer import Transformer
from .masked import Masked
from .extract import Extract



import keras
from .attention import Attention
from ..activations.gelu import gelu
from .wrapper import Wrapper


class MultiHeadAttention(Wrapper):
    """Generate multi-head attention layers.

    See: https://arxiv.org/pdf/1706.03762.pdf
    """

    def __init__(self,
                 head_num,
                 dropout_rate=0.1,
                 **kwargs):
        """Initialize the layer.

        :param head_num: Number of heads.
        :param dropout_rate: Dropout rate.
        :param feature_dim: The dimension of input feature.
        """
        self.supports_masking = True
        self.head_num = head_num
        self.dropout_rate = dropout_rate
        super(MultiHeadAttention, self).__init__(**kwargs)

    def get_config(self):
        config = {
            'head_num': self.head_num,
            'dropout_rate': self.dropout_rate,
        }
        base_config = super(MultiHeadAttention, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape

    def compute_mask(self, inputs, input_mask=None):
        return input_mask

    def build(self, input_shape):
        feature_dim = input_shape[-1]
        if feature_dim % self.head_num != 0:
            raise IndexError('Invalid head number %d with the given input dim %d' % (self.head_num, feature_dim))
        for i in range(self.head_num):
            layer = keras.layers.Dense(
                units=feature_dim // self.head_num,
                activation=gelu,
                use_bias=False,
                trainable=self.trainable,
                name='%s-Dense-Q_%d' % (self.name, i + 1),
            )
            self.layers[layer.name] = layer
            layer = keras.layers.Dense(
                units=feature_dim // self.head_num,
                activation=gelu,
                use_bias=False,
                trainable=self.trainable,
                name='%s-Dense-K_%d' % (self.name, i + 1),
            )
            self.layers[layer.name] = layer
            layer = keras.layers.Dense(
                units=feature_dim // self.head_num,
                activation=gelu,
                use_bias=False,
                trainable=self.trainable,
                name='%s-Dense-V_%d' % (self.name, i + 1),
            )
            self.layers[layer.name] = layer
            layer = keras.layers.Dropout(
                rate=self.dropout_rate,
                trainable=self.trainable,
                name='%s-Dense-Dropout-Q_%d' % (self.name, i + 1),
            )
            self.layers[layer.name] = layer
            layer = keras.layers.Dropout(
                rate=self.dropout_rate,
                trainable=self.trainable,
                name='%s-Dense-Dropout-K_%d' % (self.name, i + 1),
            )
            self.layers[layer.name] = layer
            layer = keras.layers.Dropout(
                rate=self.dropout_rate,
                trainable=self.trainable,
                name='%s-Dense-Dropout-V_%d' % (self.name, i + 1),
            )
            self.layers[layer.name] = layer
            layer = Attention(
                trainable=self.trainable,
                name='%s-Attention_%d' % (self.name, i + 1),
            )
            self.layers[layer.name] = layer
            layer = keras.layers.Dropout(
                rate=self.dropout_rate,
                trainable=self.trainable,
                name='%s-Attention-Dropout_%d' % (self.name, i + 1),
            )
            self.layers[layer.name] = layer
        if self.head_num > 1:
            layer = keras.layers.Concatenate(name='%s-Concat' % self.name)
            self.layers[layer.name] = layer
        layer = keras.layers.Dense(
            units=feature_dim,
            activation=gelu,
            use_bias=False,
            trainable=self.trainable,
            name='%s-Dense_O' % self.name,
        )
        self.layers[layer.name] = layer
        layer = keras.layers.Dropout(
            rate=self.dropout_rate,
            trainable=self.trainable,
            name='%s-Dense-Dropout_O' % self.name,
        )
        self.layers[layer.name] = layer
        super(MultiHeadAttention, self).build(input_shape)

    def call(self, inputs, mask=None):
        outputs = []
        for i in range(self.head_num):
            query_layer = self.layers['%s-Dense-Q_%d' % (self.name, i + 1)](inputs)
            key_layer = self.layers['%s-Dense-K_%d' % (self.name, i + 1)](inputs)
            value_layer = self.layers['%s-Dense-V_%d' % (self.name, i + 1)](inputs)
            query_dropout_layer = self.layers['%s-Dense-Dropout-Q_%d' % (self.name, i + 1)](query_layer)
            key_dropout_layer = self.layers['%s-Dense-Dropout-K_%d' % (self.name, i + 1)](key_layer)
            value_dropout_layer = self.layers['%s-Dense-Dropout-V_%d' % (self.name, i + 1)](value_layer)
            att_layer = self.layers['%s-Attention_%d' % (self.name, i + 1)]([
                query_dropout_layer,
                key_dropout_layer,
                value_dropout_layer,
            ])
            dropout_layer = self.layers['%s-Attention-Dropout_%d' % (self.name, i + 1)](att_layer)
            outputs.append(dropout_layer)
        if self.head_num == 1:
            concat_layer = outputs[0]
        else:
            concat_layer = self.layers['%s-Concat' % self.name](outputs)
        dense_layer = self.layers['%s-Dense_O' % self.name](concat_layer)
        dropout_layer = self.layers['%s-Dense-Dropout_O' % self.name](dense_layer)
        return dropout_layer



import unittest
import keras
import numpy as np
from keras_bert.layers import LayerNormalization, MultiHeadAttention


class TestLayerNorm(unittest.TestCase):

    def test_sample(self):
        input_layer = keras.layers.Input(
            shape=(2, 3),
            name='Input',
        )
        norm_layer = LayerNormalization(
            name='Layer-Normalization',
        )(input_layer)
        model = keras.models.Model(
            inputs=input_layer,
            outputs=norm_layer,
        )
        model.compile(
            optimizer='adam',
            loss='mse',
            metrics=['mse'],
        )
        model.summary()
        inputs = np.array([[
            [0.2, 0.1, 0.3],
            [0.5, 0.1, 0.1],
        ]])
        predict = model.predict(inputs)
        expected = np.asarray([[
            [0.0, -1.22474487, 1.22474487],
            [1.41421356, -0.707106781, -0.707106781],
        ]])
        self.assertTrue(np.allclose(expected, predict))

    def test_fit(self):
        input_layer = keras.layers.Input(
            shape=(2, 3),
            name='Input',
        )
        att_layer = MultiHeadAttention(
            head_num=3,
            dropout_rate=1e-5,
            name='MH'
        )(input_layer)
        dense_layer = keras.layers.Dense(units=3, name='Dense-1')(att_layer)
        norm_layer = LayerNormalization(
            name='Layer-Normalization',
            trainable=False,
        )(dense_layer)
        dense_layer = keras.layers.Dense(units=3, name='Dense-2')(norm_layer)
        model = keras.models.Model(
            inputs=input_layer,
            outputs=dense_layer,
        )
        model.compile(
            optimizer=keras.optimizers.Adam(lr=1e-3),
            loss='mse',
            metrics=['mse'],
        )
        model.summary()

        def _generator(batch_size=32):
            while True:
                inputs = np.random.random((batch_size, 2, 3))
                outputs = np.asarray([[[0.0, -0.1, 0.2]] * 2] * batch_size)
                yield inputs, outputs

        model.fit_generator(
            generator=_generator(),
            steps_per_epoch=1000,
            epochs=30,
            validation_data=_generator(),
            validation_steps=100,
            callbacks=[
                keras.callbacks.EarlyStopping(monitor='val_loss', patience=5)
            ],
        )
        for inputs, _ in _generator(batch_size=3):
            predicts = model.predict(inputs)
            expect = np.round(np.asarray([[[0.0, -0.1, 0.2]] * 2] * 3), decimals=1)
            actual = np.round(predicts, decimals=1)
            self.assertTrue(np.allclose(expect, actual), (expect, actual))
            break



import unittest
import random
import keras
import numpy as np
from keras_bert.layers import get_inputs, Embeddings, Transformer, Masked


class TestMasked(unittest.TestCase):

    def test_sample(self):
        inputs = get_inputs(seq_len=512)
        embed_layer = Embeddings(input_dim=12, output_dim=768, position_dim=512)(inputs)
        masked_layer = Masked(name='Masked')([embed_layer, inputs[-1]])
        model = keras.models.Model(inputs=inputs, outputs=masked_layer)
        model.compile(
            optimizer='adam',
            loss='mse',
            metrics=['mse'],
        )
        model.summary(line_length=120)
        model.predict([
            np.asarray([[1] + [0] * 511]),
            np.asarray([[0] * 512]),
            np.asarray([[0] * 512]),
            np.asarray([[1] + [0] * 511]),
        ])
        self.assertEqual((None, 512, 768), model.layers[-1].output_shape)

    def test_mask_result(self):
        input_layer = keras.layers.Input(
            shape=(None,),
            name='Input',
        )
        embed_layer = keras.layers.Embedding(
            input_dim=12,
            output_dim=9,
            mask_zero=True,
            name='Embedding',
        )(input_layer)
        transformer_layer = Transformer(
            head_num=1,
            hidden_dim=12,
            dropout_rate=0.1,
            name='Transformer',
        )(embed_layer)
        dense_layer = keras.layers.Dense(
            units=12,
            activation='softmax',
            name='Dense',
        )(transformer_layer)
        mask_layer = keras.layers.Input(
            shape=(None,),
            name='Mask',
        )
        masked_layer, mask_result = Masked(
            return_masked=True,
            name='Masked',
        )([dense_layer, mask_layer])
        model = keras.models.Model(
            inputs=[input_layer, mask_layer],
            outputs=[masked_layer, mask_result],
        )
        model.compile(
            optimizer='adam',
            loss='mse',
            metrics=['mse'],
        )
        model.summary(line_length=150)
        predicts = model.predict([
            np.asarray([
                [1, 2, 3, 4, 5, 6, 7, 8, 0, 0],
                [1, 2, 3, 4, 0, 0, 0, 0, 0, 0],
            ]),
            np.asarray([
                [0, 0, 1, 0, 0, 0, 1, 0, 0, 0],
                [0, 1, 0, 1, 0, 0, 0, 0, 0, 0],
            ]),
        ])
        expect = np.asarray([
            [0, 0, 1, 0, 0, 0, 1, 0, 0, 0],
            [0, 1, 0, 1, 0, 0, 0, 0, 0, 0],
        ])
        self.assertTrue(np.allclose(expect, predicts[1]))

    def test_fit(self):
        input_layer = keras.layers.Input(
            shape=(15,),
            name='Input',
        )
        embed_layer = keras.layers.Embedding(
            input_dim=12,
            output_dim=24,
            mask_zero=True,
            name='Embedding',
        )(input_layer)
        rnn_layer = keras.layers.Bidirectional(
            keras.layers.LSTM(units=100, return_sequences=True),
            name='Bi-LSTM',
        )(embed_layer)
        dense_layer = keras.layers.Dense(
            units=12,
            activation='softmax',
            name='Dense',
        )(rnn_layer)
        mask_layer = keras.layers.Input(
            shape=(None,),
            name='Mask',
        )
        masked_layer = Masked(
            name='Masked',
        )([dense_layer, mask_layer])
        model = keras.models.Model(
            inputs=[input_layer, mask_layer],
            outputs=masked_layer,
        )
        model.compile(
            optimizer=keras.optimizers.Adam(lr=1e-4),
            loss=keras.losses.sparse_categorical_crossentropy,
            metrics=[keras.metrics.sparse_categorical_crossentropy],
        )
        model.summary(line_length=150)

        def _generator(batch_size=32):
            while True:
                inputs, masked, outputs = [], [], []
                for _ in range(batch_size):
                    inputs.append([])
                    masked.append([])
                    outputs.append([])
                    has_mask = False
                    for i in range(1, 11):
                        inputs[-1].append(i)
                        outputs[-1].append([i])
                        if random.random() < 0.3:
                            has_mask = True
                            inputs[-1][-1] = 11
                            masked[-1].append(1)
                        else:
                            masked[-1].append(0)
                    if not has_mask:
                        masked[-1][0] = 1
                    inputs[-1] += [0] * (15 - len(inputs[-1]))
                    masked[-1] += [0] * (15 - len(masked[-1]))
                    outputs[-1] += [[0]] * (15 - len(outputs[-1]))
                yield [np.asarray(inputs), np.asarray(masked)], np.asarray(outputs)

        model.fit_generator(
            generator=_generator(),
            steps_per_epoch=1000,
            epochs=10,
            validation_data=_generator(),
            validation_steps=100,
            callbacks=[
                keras.callbacks.EarlyStopping(monitor='val_loss', patience=2)
            ],
        )
        for inputs, outputs in _generator(batch_size=32):
            predicts = model.predict(inputs)
            actual = np.argmax(predicts, axis=-1)
            for i in range(32):
                for j in range(15):
                    if inputs[1][i][j]:
                        self.assertEqual(j + 1, actual[i][j])
            break



import unittest
import keras
import numpy as np
from keras_bert.layers import MultiHeadAttention


class TestMultiHead(unittest.TestCase):

    def test_sample(self):
        input_layer = keras.layers.Input(
            shape=(512,),
            name='Input',
        )
        embed_layer = keras.layers.Embedding(
            input_dim=12,
            output_dim=768,
            mask_zero=True,
            name='Embedding',
        )(input_layer)
        output_layer = MultiHeadAttention(
            head_num=12,
            name='Multi-Head',
        )(embed_layer)
        model = keras.models.Model(inputs=input_layer, outputs=output_layer)
        model.compile(
            optimizer='adam',
            loss='mse',
            metrics=['mse'],
        )
        model.summary()
        self.assertEqual((None, 512, 768), model.layers[-1].output_shape)

    def test_invalid_head_num(self):
        with self.assertRaises(IndexError):
            input_layer = keras.layers.Input(
                shape=(2, 3),
                name='Input',
            )
            MultiHeadAttention(
                head_num=2,
                dropout_rate=0.01,
                name='Multi-Head',
            )(input_layer)

    def test_fit(self):
        input_layer = keras.layers.Input(
            shape=(2, 3),
            name='Input',
        )
        att_layer = MultiHeadAttention(
            head_num=3,
            dropout_rate=0.01,
            name='Multi-Head-1',
        )(input_layer)
        dense_layer = keras.layers.Dense(units=3, name='Dense-1')(att_layer)
        att_layer = MultiHeadAttention(
            head_num=3,
            name='Multi-Head-2',
        )(dense_layer)
        output_layer = keras.layers.Dense(units=3, name='Dense-2')(att_layer)
        model = keras.models.Model(inputs=input_layer, outputs=output_layer)
        model.compile(
            optimizer='adam',
            loss='mse',
            metrics=['mse'],
        )
        model.summary()

        def _generator(batch_size=32):
            while True:
                inputs = np.random.random((batch_size, 2, 3))
                outputs = np.asarray([[[0.0, -0.1, 0.2]] * 2] * batch_size)
                yield inputs, outputs

        model.fit_generator(
            generator=_generator(),
            steps_per_epoch=1000,
            epochs=10,
            validation_data=_generator(),
            validation_steps=100,
            callbacks=[
                keras.callbacks.EarlyStopping(monitor='val_loss', patience=5)
            ],
        )
        for inputs, _ in _generator(batch_size=3):
            predicts = model.predict(inputs)
            expect = np.asarray([[[0.0, -0.1, 0.2]] * 2] * 3)
            actual = np.round(predicts, decimals=1)
            self.assertTrue(np.allclose(expect, actual), (expect, actual))
            break



import keras
import keras.backend as K


class LayerNormalization(keras.layers.Layer):
    """Layer normalization.

    See: https://arxiv.org/pdf/1607.06450.pdf
    """

    def __init__(self, **kwargs):
        self.supports_masking = True
        self.gamma, self.beta = None, None
        super(LayerNormalization, self).__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        return input_shape

    def compute_mask(self, inputs, input_mask=None):
        return input_mask

    def build(self, input_shape):
        self.gamma = self.add_weight(shape=input_shape[-1:],
                                     name='{}_gamma'.format(self.name),
                                     initializer=keras.initializers.get('ones'))
        self.beta = self.add_weight(shape=input_shape[-1:],
                                    name='{}_beta'.format(self.name),
                                    initializer=keras.initializers.get('zeros'))
        super(LayerNormalization, self).build(input_shape)

    def call(self, x, mask=None):
        mean = K.mean(x, axis=-1, keepdims=True)
        std = K.std(x, axis=-1, keepdims=True)
        return self.gamma * (x - mean) / (std + K.epsilon()) + self.beta



import keras


class Wrapper(keras.layers.Layer):

    CONFIG_PREFIX = 'wrapper_layer_'

    def __init__(self, layers=None, **kwargs):
        if layers is None:
            self.layers = {}
        else:
            self.layers = layers
            self.built = True
        super(Wrapper, self).__init__(**kwargs)

    @property
    def trainable_weights(self):
        weights = []
        for key in sorted(self.layers.keys()):
            weights += self.layers[key].trainable_weights
        return weights

    @property
    def non_trainable_weights(self):
        weights = []
        for key in sorted(self.layers.keys()):
            weights += self.layers[key].non_trainable_weights
        return weights

    def get_config(self):
        config = {}
        for name, layer in self.layers.items():
            config[self.CONFIG_PREFIX + name] = {
                'class_name': layer.__class__.__name__,
                'config': layer.get_config(),
            }
        base_config = super(Wrapper, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @classmethod
    def from_config(cls, config, custom_objects=None):
        layers = {}
        keys = list(filter(lambda key: key.startswith(Wrapper.CONFIG_PREFIX), config.keys()))
        for key in keys:
            if key.startswith(Wrapper.CONFIG_PREFIX):
                name = key[len(Wrapper.CONFIG_PREFIX):]
                layers[name] = keras.layers.deserialize(config.pop(key), custom_objects=custom_objects)
        return cls(layers=layers, **config)



import unittest
import keras
import numpy as np
from keras_bert.layers import FeedForward


class TestAttention(unittest.TestCase):

    def test_sample(self):
        input_layer = keras.layers.Input(
            shape=(1, 3),
            name='Input',
        )
        feed_forward_layer = FeedForward(
            hidden_dim=4,
            weights=[
                np.asarray([
                    [0.1, 0.2, 0.3, 0.4],
                    [-0.1, 0.2, -0.3, 0.4],
                    [0.1, -0.2, 0.3, -0.4],
                ]),
                np.asarray([
                    0.0, -0.1, 0.2, -0.3,
                ]),
                np.asarray([
                    [0.1, 0.2, 0.3],
                    [-0.1, 0.2, -0.3],
                    [0.1, -0.2, 0.3],
                    [-0.1, 0.2, 0.3],
                ]),
                np.asarray([
                    0.0, 0.1, -0.2,
                ]),
            ],
            name='FeedForward',
        )(input_layer)
        model = keras.models.Model(
            inputs=input_layer,
            outputs=feed_forward_layer,
        )
        model.compile(
            optimizer='adam',
            loss='mse',
            metrics=['mse'],
        )
        model.summary()
        inputs = np.array([[[0.2, 0.1, 0.3]]])
        predict = model.predict(inputs)
        expected = np.asarray([[[0.03814463, 0.03196599, -0.15434355]]])
        self.assertTrue(np.allclose(expected, predict), predict)

    def test_fit(self):
        input_layer = keras.layers.Input(
            shape=(1, 3),
            name='Input',
        )
        feed_forward_layer = FeedForward(
            hidden_dim=4,
            name='FeedForward',
        )(input_layer)
        model = keras.models.Model(
            inputs=input_layer,
            outputs=feed_forward_layer,
        )
        model.compile(
            optimizer='adam',
            loss='mse',
            metrics=['mse'],
        )

        def _generator(batch_size=32):
            while True:
                inputs = np.random.random((batch_size, 1, 3))
                outputs = inputs * 0.8 + 0.3
                yield inputs, outputs

        for _ in range(3):
            model.fit_generator(
                generator=_generator(),
                steps_per_epoch=1000,
                epochs=30,
                validation_data=_generator(),
                validation_steps=100,
                callbacks=[
                    keras.callbacks.EarlyStopping(monitor='val_loss', patience=5)
                ],
            )
            for inputs, _ in _generator(batch_size=3):
                predicts = model.predict(inputs)
                expect = np.round(inputs * 0.8 + 0.3, decimals=1)
                actual = np.round(predicts, decimals=1)
                if np.allclose(expect, actual):
                    return
                break
        self.assertTrue(np.allclose(expect, actual), (expect, actual))



import unittest
import keras
import numpy as np
from keras_bert.layers import Attention


class TestAttention(unittest.TestCase):

    def test_sample(self):
        input_layer = keras.layers.Input(
            shape=(5,),
            name='Input',
        )
        embed_layer = keras.layers.Embedding(
            input_dim=4,
            output_dim=5,
            mask_zero=True,
            weights=[
                np.asarray([
                    [0.1, 0.2, 0.3, 0.4, 0.5],
                    [0.2, 0.3, 0.4, 0.6, 0.5],
                    [0.4, 0.7, 0.2, 0.6, 0.9],
                    [0.3, 0.5, 0.8, 0.9, 0.1],
                ]),
            ],
            name='Embedding',
        )(input_layer)
        att_layer = Attention(
            name='Attention',
        )([embed_layer, embed_layer, embed_layer])
        model = keras.models.Model(inputs=input_layer, outputs=att_layer)
        model.compile(
            optimizer='adam',
            loss='mse',
            metrics=['mse'],
        )
        model.summary()
        inputs = np.array([[1, 2, 3, 1, 0]])
        predict = model.predict(inputs)[0]
        self.assertTrue(np.allclose(predict[0], predict[3]))
        self.assertTrue(np.allclose(
            np.asarray([0.24973875, 0.41574854, 0.44609764, 0.644485, 0.47796533]),
            predict[2],
        ), predict[2])



import keras
import keras.backend as K


class Attention(keras.layers.Layer):
    """Attention layer.

    See: https://arxiv.org/pdf/1706.03762.pdf
    """

    def __init__(self, **kwargs):
        self.supports_masking = True
        super(Attention, self).__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        return input_shape[-1]

    def compute_mask(self, inputs, mask=None):
        if isinstance(mask, list):
            return mask[-1]
        return mask

    def call(self, inputs, mask=None, **kwargs):
        query, key, value = inputs
        feature_dim = K.shape(query)[-1]
        e = K.batch_dot(query, key, axes=2) / K.sqrt(K.cast(feature_dim, dtype=K.floatx()))
        if isinstance(mask, list) and mask[-1] is not None:
            e -= (1.0 - K.cast(K.expand_dims(mask[-1]), K.floatx())) * 1e9
        a = keras.activations.softmax(e)
        v = K.batch_dot(a, value)
        return v



import keras
import keras.backend as K
from ..activations import gelu


class FeedForward(keras.layers.Layer):
    """Position-wise feed-forward layer.

    See: https://arxiv.org/pdf/1706.03762.pdf
    """

    def __init__(self, hidden_dim, **kwargs):
        self.supports_masking = True
        self.hidden_dim = hidden_dim
        self.W1, self.b1 = None, None
        self.W2, self.b2 = None, None
        super(FeedForward, self).__init__(**kwargs)

    def get_config(self):
        config = {
            'hidden_dim': self.hidden_dim,
        }
        base_config = super(FeedForward, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape

    def compute_mask(self, inputs, input_mask=None):
        return input_mask

    def build(self, input_shape):
        feature_dim = input_shape[-1]
        self.W1 = self.add_weight(shape=(feature_dim, self.hidden_dim),
                                  name='{}_W1'.format(self.name),
                                  initializer=keras.initializers.get('glorot_normal'))
        self.b1 = self.add_weight(shape=(self.hidden_dim,),
                                  name='{}_b1'.format(self.name),
                                  initializer=keras.initializers.get('zeros'))
        self.W2 = self.add_weight(shape=(self.hidden_dim, feature_dim),
                                  name='{}_W2'.format(self.name),
                                  initializer=keras.initializers.get('glorot_normal'))
        self.b2 = self.add_weight(shape=(feature_dim,),
                                  name='{}_b2'.format(self.name),
                                  initializer=keras.initializers.get('zeros'))
        super(FeedForward, self).build(input_shape)

    def call(self, x, mask=None):
        return K.dot(gelu(K.dot(x, self.W1) + self.b1), self.W2) + self.b2



import keras
from .multi_head import MultiHeadAttention
from .layer_norm import LayerNormalization
from .feed_forward import FeedForward
from .wrapper import Wrapper


class Transformer(Wrapper):
    """Generate a set of transformer layers.

    See: https://arxiv.org/pdf/1706.03762.pdf
    """

    def __init__(self,
                 head_num,
                 hidden_dim,
                 dropout_rate=0.1,
                 **kwargs):
        """Initialize the layer.

        :param head_num: Number of heads.
        :param hidden_dim: Hidden dimension for feed forward layer.
        :param dropout_rate: Dropout rate.
        """
        self.supports_masking = True
        self.head_num = head_num
        self.hidden_dim = hidden_dim
        self.dropout_rate = dropout_rate
        super(Transformer, self).__init__(**kwargs)

    def get_config(self):
        config = {
            'head_num': self.head_num,
            'hidden_dim': self.hidden_dim,
            'dropout_rate': self.dropout_rate,
        }
        base_config = super(Transformer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape

    def compute_mask(self, inputs, input_mask=None):
        return input_mask

    def build(self, input_shape):
        layer = MultiHeadAttention(
            head_num=self.head_num,
            dropout_rate=self.dropout_rate,
            trainable=self.trainable,
            name='%s-MultiHead' % self.name,
        )
        self.layers[layer.name] = layer
        layer = LayerNormalization(
            trainable=self.trainable,
            name='%s-MultiHead-Norm' % self.name,
        )
        self.layers[layer.name] = layer
        layer = keras.layers.Dropout(
            rate=self.dropout_rate,
            trainable=self.trainable,
            name='%s-MultiHead-Dropout' % self.name,
        )
        self.layers[layer.name] = layer
        layer = keras.layers.Add(
            trainable=self.trainable,
            name='%s-MultiHead-Add' % self.name,
        )
        self.layers[layer.name] = layer
        layer = FeedForward(
            hidden_dim=self.hidden_dim,
            trainable=self.trainable,
            name='%s-FeedForward' % self.name,
        )
        self.layers[layer.name] = layer
        layer = LayerNormalization(
            trainable=self.trainable,
            name='%s-FeedForward-Norm' % self.name,
        )
        self.layers[layer.name] = layer
        layer = keras.layers.Dropout(
            rate=self.dropout_rate,
            trainable=self.trainable,
            name='%s-FeedForward-Dropout' % self.name,
        )
        self.layers[layer.name] = layer
        layer = keras.layers.Add(
            trainable=self.trainable,
            name='%s-FeedForward-Add' % self.name,
        )
        self.layers[layer.name] = layer
        super(Transformer, self).build(input_shape)

    def call(self, inputs, mask=None):
        multi_head_layer = self.layers['%s-MultiHead' % self.name](inputs)
        multi_head_norm = self.layers['%s-MultiHead-Norm' % self.name](multi_head_layer)
        multi_head_dropout_layer = self.layers['%s-MultiHead-Dropout' % self.name](multi_head_norm)
        multi_head_residual_layer = self.layers['%s-MultiHead-Add' % self.name]([inputs, multi_head_dropout_layer])
        feed_forward_layer = self.layers['%s-FeedForward' % self.name](multi_head_residual_layer)
        feed_forward_norm = self.layers['%s-FeedForward-Norm' % self.name](feed_forward_layer)
        feed_forward_dropout_layer = self.layers['%s-FeedForward-Dropout' % self.name](feed_forward_norm)
        feed_forward_residual_layer = self.layers['%s-FeedForward-Add' % self.name](
            [multi_head_residual_layer, feed_forward_dropout_layer]
        )
        return feed_forward_residual_layer
